@@ -1,3 +1,8 @@
+{-# LANGUAGE CPP #-}
+#if __GLASGOW_HASKELL__ >= 702
+{-# LANGUAGE DeriveGeneric #-}
+#endif
+
 -- (c) 1999 - 2002 by Martin Erwig [see file COPYRIGHT]
 -- | Tree-based implementation of 'Graph' and 'DynGraph'
 --
@@ -9,17 +14,28 @@ module Data.Graph.Inductive.Tree (Gr,UGr) where
 import Data.Graph.Inductive.Graph
 
 import           Control.Applicative (liftA2)
-import           Control.Arrow       (first)
+import           Control.Arrow       (first, second)
 import           Data.List           (foldl', sort)
 import           Data.Map            (Map)
 import qualified Data.Map            as M
 import           Data.Maybe          (fromMaybe)
 
+#if __GLASGOW_HASKELL__ >= 704
+import Control.DeepSeq (NFData (..))
+#endif
+
+#if __GLASGOW_HASKELL__ >= 702
+import GHC.Generics (Generic)
+#endif
+
 ----------------------------------------------------------------------
 -- GRAPH REPRESENTATION
 ----------------------------------------------------------------------
 
 newtype Gr a b = Gr (GraphRep a b)
+#if __GLASGOW_HASKELL__ >= 702
+  deriving (Generic)
+#endif
 
 type GraphRep a b = Map Node (Context' a b)
 type Context' a b = (Adj b,a,Adj b)
@@ -33,7 +49,7 @@ type UGr = Gr () ()
 instance (Eq a, Ord b) => Eq (Gr a b) where
   (Gr g1) == (Gr g2) = fmap sortAdj g1 == fmap sortAdj g2
     where
-      sortAdj (a1,n,a2) = (sort a1,n,sort a2)
+      sortAdj (p,n,s) = (sort p,n,sort s)
 
 instance (Show a, Show b) => Show (Gr a b) where
   showsPrec d g = showParen (d > 10) $
@@ -53,27 +69,34 @@ instance (Read a, Read b) => Read (Gr a b) where
 --
 instance Graph Gr where
   empty             = Gr M.empty
+
   isEmpty (Gr g)    = M.null g
+
   match v gr@(Gr g) = maybe (Nothing, gr)
                             (first Just . uncurry (cleanSplit v))
                       . (\(m,g') -> fmap (flip (,) g') m)
                       $ M.updateLookupWithKey (const (const Nothing)) v g
-  mkGraph vs es     = (insEdges' . insNodes vs) empty
-        where
-          insEdges' g = foldl' (flip insEdge) g es
+
+  mkGraph vs es     = insEdges es
+                      . Gr
+                      . M.fromList
+                      . map (second (\l -> ([],l,[])))
+                      $ vs
 
   labNodes (Gr g)   = map (\(v,(_,l,_))->(v,l)) (M.toList g)
-  -- more efficient versions of derived class members
-  --
+
   matchAny (Gr g)   = maybe (error "Match Exception, Empty Graph")
                             (uncurry (uncurry cleanSplit))
                             (M.minViewWithKey g)
+
   noNodes   (Gr g)  = M.size g
-  nodeRange (Gr g)  = fromMaybe (0,0)
+
+  nodeRange (Gr g)  = fromMaybe (error "nodeRange of empty graph")
                       $ liftA2 (,) (ix (M.minViewWithKey g))
                                    (ix (M.maxViewWithKey g))
     where
       ix            = fmap (fst . fst)
+
   labEdges  (Gr g)  = concatMap (\(v,(_,_,s))->map (\(l,w)->(v,w,l)) s) (M.toList g)
 
 -- After a Node (with its corresponding Context') are split out of a
@@ -102,6 +125,11 @@ instance DynGraph Gr where
                        (const (error ("Node Exception, Node: "++show v)))
       cntxt' = (p,l,s)
 
+#if __GLASGOW_HASKELL__ >= 704
+instance (NFData a, NFData b) => NFData (Gr a b) where
+  rnf (Gr g) = rnf g
+#endif
+
 ----------------------------------------------------------------------
 -- UTILITIES
 ----------------------------------------------------------------------
